package se.ericsson.eto.norarc.javaframe;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * A composite of ActiveObjects collected in a list.
 * A Composite will normally have Mediators on its edges.
 * @see Mediator
 * @see ActiveObject
 */
public class Composite extends ActiveObject {
  public LinkedList activeObjectList;

  /**
   * The default constructor has no parameters, but any given specific
   * extension's constructor will normally have the edge {@link Mediator}s as parameters
   */
  public Composite() {
    activeObjectList = new LinkedList();
  }


  /**
   * Add an {@link ActiveObject} to the composite
   */
  public void addActiveObject(ActiveObject newActiveObject) {
    newActiveObject.owner = this;
    activeObjectList.add(newActiveObject);
  }

  /**
   * Removes oldActiveObject from the composite's containment list and 
   * removes all links to the mediators of oldActiveObject.<br>
   * <br>
   * For this to work two preconditions must be met:<br>
   * 1. All formal and actual mediators in the class of the composite object 
   * (i.e., the class of this object) must be be declared as public fields.<br>
   * 2. All formal mediators in the class of oldActiveObject must be declared
   * as public fields.
   * 
   * @param oldActiveObject - the activeObject to be removed
   */
	public void removeActiveObject(ActiveObject oldActiveObject){
		/* Remove oldActiveObject from the composite's list */
		activeObjectList.remove(oldActiveObject);
		
		try {
			/* Get all mediator objects assigned to fields declared by oldActiveObject's class 
			 * Should later remove all links to these mediator objects */
			Class mediatorClass = Class.forName("se.ericsson.eto.norarc.javaframe.Mediator");
			List addressesToRemove = new ArrayList();
			Field[] f = oldActiveObject.getClass().getDeclaredFields();
			for (int i = 0; i < f.length; i++) {
				// The field must be of type Mediator or a subtype and it must be public
				if (mediatorClass.isAssignableFrom(f[i].getType()) && Modifier.isPublic(f[i].getModifiers())) {
					Mediator m = (Mediator)f[i].get(oldActiveObject);
					addressesToRemove.add(m);
				}
			}
	
			/* We now have a list of mediator objects that we need to remove links to. 
			 * Gets all mediator objects assigned to fields declared in this composite class
			 * and removes all links from them to the mediator objects of oldActiveObject */
			Field[] fields = this.getClass().getDeclaredFields();
			for (int i = 0; i < fields.length; i++) {
				if (mediatorClass.isAssignableFrom(fields[i].getType()) && Modifier.isPublic(fields[i].getModifiers())) {
					Mediator m = (Mediator)fields[i].get(this);
					for (Iterator it = addressesToRemove.iterator(); it.hasNext();) {
						Mediator med = (Mediator) it.next();
						m.removeAddress(med);
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
