package se.ericsson.eto.norarc.javaframe;

/**
 * CompositeState is a {@link State} with inner States.
 * A CompositeState has 1 or more entry and exit ports that connects
 * the CompositeState to its environment.
 * The ports are identified by an entry index or an exit index (int).
 * A CompositeState may have a default entry port (not identified by an index).
 */
public abstract class CompositeState extends State {

  /**
   * CompositeState constructor with user supplied name.
   * @param sn The name of this CompositeState.
   */
  public CompositeState(String sn) {
    super(sn);
  }

  /**
   * CompositeState constructor with default name.
   */
  public CompositeState() {
    super();
  }

  // ####################################################################
  // Utilities to be used within the transitions, that is inside the
  // methods enterState(), execTrans(), and outofInnerCompositeState()
  // (see below).

  /**
   * Leave this CompositeState through a given exit port.
   * The method is to be used within the methods execTrans() or outofInnerCompositeState().
   * @param exitNo Index of the exit port.
   * @param curfsm The current StateMachine.
   */
  protected final void exitState(int exitNo, StateMachine curfsm) {
    exit(curfsm);
    curfsm.currentState = enclosingState;
    enclosingState.outofInnerCompositeState(this, exitNo, curfsm);
  }

  /**
   * Execute exit() from currentState up to (but not including) this CompositeState.
   * The method is to be used within the method execTrans().
   * @param curfsm The current StateMachine.
   */
  protected final void performExit(StateMachine curfsm) {
    State st = curfsm.currentState;
    do {
      st.exit(curfsm);
      st = st.enclosingState;
    } while (this != st);
    curfsm.currentState = this;
  }

  /**
   * Remain in the same State of the StateMachine.
   * The method is to be used within the method execTrans().
   * Notice that entry() of this State will not be executed. That means
   * performExit() should neither have been called.
   * @param curfsm  The current StateMachine.
   */
  protected final void sameState(StateMachine curfsm) {
    if (curfsm.myScheduler.trc.trace_on) // Trace the end of the transition
      curfsm.myScheduler.trc.traceSameState(curfsm.currentState);
  }

  /**
   * The received Message is not to be consumed and handled in this State
   * and the Message is saved in order to be handled later in another State.
   * The method is to be used within the method execTrans().
   * @param sig    The Message that will be saved (assumed to be != null).
   * @param curfsm The StateMachine that conceptually performs this operation
   *               (assumed to be != null).
   */
  protected final void save(Message sig, StateMachine curfsm) {
    curfsm.saveQueue.addMessage(sig); // put signal into the save queue
    if (curfsm.myScheduler.trc.trace_on) // Trace the end of the transition
      curfsm.myScheduler.trc.traceSave(curfsm, curfsm.currentState, sig);
  }



  // ####################################################################
  // Methods that has to be written for each non-abstract specialization
  // of CompositeState.

  /**
   * Enter this CompositeState through a given entry port.
   * enterState links the given entry port with an inner State
   * based on the value of the entry index. It may also contain optional
   * user transition code.
   * Note: enterState has two versions
   * <br>
   *   - with entry port index
   * <br>
   *   - without entry port index (entry through the default entry port).
   * <br>
   * For each CompositeState class at least one of the enterState methods
   * has to be redefined.
   * NB: enterState is called from user transition code.
   * @param enterNo Index of the entry port.
   * @param curfsm The current StateMachine.
   */
  public void enterState(int enterNo, StateMachine curfsm) {
    System.exit(1); // should not be called if not redefined
  }

  /**
   * Enter this CompositeState through the default entry port.
   * enterState links the default entry port with an inner State.
   * It may also contain optional user transition code.
   * NB: enterState is called from user transition code.
   * @param curfsm The current StateMachine.
   */
  public void enterState(StateMachine curfsm) {
    System.exit(1); // should not be called if not redefined
  }

  /**
   * Execute transition (if defined) for Message sig and State st.
   * For each CompositeState class execTrans has to be defined.
   * It defines the transitions (and saves) handled by its inner States
   * (not included those inside inner CompositeStates).
   * NB: execTrans is not called from user code.
   * @param sig The consumed Message.
   * @param st  The State to search for defined transition.
   * @param curfsm The performing StateMachine.
   * @return If a transition has been executed (or save), then return true.
   */
  protected abstract boolean execTrans(Message sig, State st, StateMachine curfsm);

  // ####################################################################
  // Method that must be written if a non-abstract CompositeState contains
  // inner CompositeState.
  /**
   * Executes linking code when an inner CompositeState leaves through a given exit port.
   * For each CompositeState class with inner CompositeStates the operation
   * outofInnerCompositeState has to be defined.
   * It will link the given exit port with an inner singular State,
   * the entry port of an inner CompositeState, or an exit port
   * of this CompositeState.
   * NB: outofInnerCompositeState is not called from user code.
   * @param cs The inner CompositeState that is leaving.
   * @param exNo Index of the exit port of the inner CompositeState.
   * @param curfsm The current StateMachine.
   */
  protected void outofInnerCompositeState(CompositeState cs, int exNo, StateMachine curfsm) {
  }
}
