

/**
 * Title:        Message GUI<p>
 * Description:  GUI for message input to JavaFrame application<p>
 * Copyright:    Copyright (c) Lars Stuevold<p>
 * Company:      Ericsson<p>
 * @author       Lars Stuevold
 * @version      1.0
 */
package se.ericsson.eto.norarc.javaframe;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.util.Vector;
import java.lang.reflect.*;

/**
 * Creates a user interface to input messages to a JavaFrame application.
 * This class is a ready-made GUI, but have to be extended. The new extended class (e.g. PtnGUI)
 * should have the following content:
 * <ol>
 *
 * <li>Import your message types and your datatypes in order to create messages.</li>
 *
 * <li>The <code>generateMessage</code> method must be given content by the user in order to interpret
 *    input parameters from this GUI. The necessary message composition must be coded before
 *    sending off the message to the selected mediator.<br>
 *    Two helper methods is provided to facilitate conversion of <code>int</code> type parameters and <code>float</code> parameters.
 *    These are <code>parToInt</code> and <code>parToFloat</code>.</li>
 *</ol>
 * These actions complete the points of extension of this class.
 *<p>
 * After this class has been created, two methods must be used to make the mediators
 * and messages selectable in the GUI.
 * To associate an alias name to a Mediator, use <code>defMediator</code>.
 * The method <code>defMessage</code> will display the message name in the GUI.<p>
 * The GUI is realized by calling <code>show</code>.
 * <br>
 * NOTE! The two methods <code>defMediator</code> and  <code>defMessage</code> must not be called
 * after <code>show</code> because <code>defMediator</code> and  <code>defMessage</code> are not thread-safe.
 *
 * However, you can use the <code>addLog</code> method to output log
 * information to the bottom window of this GUI,
 *  e.g. in the forward method of JavaFrame's output edge mediator(s) or in the <code>generateMessage</code> method
 *  of this class.
 *  Here is a code example:
 *  <code><pre>
    // generate the message GUI object
    env = new PtnGUI("Input to Pattern Composite Example");
    // inform GUI about the input message types,using class literal
    env.defMessage(PtnMessage.class);
    // or use fully qualified class name
    env.defMessage("se.ericsson.eto.norarc.ptn.PtnMessage");
    // inform GUI about the input mediator
    env.defMediator("ToSystem", inputMediator);
    // display GUI
    env.show();
   </pre></code>
 */

public abstract class MessageGUI extends JFrame{
  private JPanel contentPane;
  private JTextPane outputMessagePane;
  private Vector vectorOfMedStrings = new Vector();
  private Vector vectorOfMedRefs = new Vector();
  private Vector vectorOfMsgsStrings = new Vector();
  private Vector vectorOfFullNameMsgsStrings = new Vector();
  private JList jlistOfMediators;
  private JList jlistOfMsgs;
  private JTextField parField;
  private JComponent f1,f2,f3,f4,f5; // to set tab focus order

  DefaultStyledDocument doc = new  DefaultStyledDocument();
  MutableAttributeSet attr;
  JScrollPane outputMsgScrollPane;

  LogInterface lif = new LogInterface();
  /**
   *  Instantiates the UI object, but nothing is diplayed before <code>show()</code> is called.
   *  @param  headerText   a <code>String</code> used as a title of the <code>JFrame</code>.
   */
  public MessageGUI(String headerText) {
    this.setTitle(headerText);
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    try {
      createMainPanel();
    }
    catch(Exception e) {
      e.printStackTrace();
    }
  }

  /**
   *  Display the GUI, invoked once after all <code>defMessage</code> calls and  all <code>defMediator</code> calls.
   */
  public void show() {

    if ( vectorOfMedStrings.size() > 0 ) {
      // Put  mediator list of into JList
      jlistOfMediators.setListData(vectorOfMedStrings);
      jlistOfMediators.setSelectedIndex(0);
    } else {
      //JOptionPane.showMessageDialog(new JFrame(), "No mediators defined before show() is called.",
      //    "No mediators", JOptionPane.ERROR_MESSAGE);
    }
    if ( vectorOfMsgsStrings.size() > 0 ) {
      // Put  message list of into JList
      jlistOfMsgs.setListData(vectorOfMsgsStrings);
      jlistOfMsgs.setSelectedIndex(0);
      jlistOfMsgs.setToolTipText((String) vectorOfFullNameMsgsStrings.elementAt(jlistOfMsgs.getSelectedIndex()));
    } else {
      //JOptionPane.showMessageDialog(new JFrame(), "No messages defined before show() is called.",
      //    "No messages", JOptionPane.ERROR_MESSAGE);
    }
    pack();
    super.show();

    // Make sure that Swing Thread has realized the widgets.
    while (! super.isRootPaneCheckingEnabled()) {
      try {
        Thread.sleep(100);
      } catch (InterruptedException ie) {
        System.err.println( "Thread Exception: " + ie) ;
      }
    }
  }

  private void createMainPanel() {
    this.setSize(new Dimension(450, 450));
    contentPane = (JPanel) this.getContentPane();
    contentPane.setBorder(BorderFactory.createLineBorder(Color.black));
    contentPane.setMinimumSize(new Dimension(420, 280));
    contentPane.setPreferredSize(new Dimension(450, 400));
    JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
                                     createSuperUpperPanel(), createLowerPanel());
    splitPane.setOneTouchExpandable(true);
    splitPane.setDividerLocation(150);
    contentPane.add(splitPane);

    // Set focus order
    f1.setNextFocusableComponent(f2);
    f2.setNextFocusableComponent(f3);
    f3.setNextFocusableComponent(f4);
    f4.setNextFocusableComponent(f5);
    f5.setNextFocusableComponent(f1);
  }

  private JPanel createSuperUpperPanel() {
    JPanel superUpperPanel = new JPanel();
    superUpperPanel.setLayout(new BorderLayout(5,5));
    superUpperPanel.add(createUpperPane(),BorderLayout.CENTER);
    superUpperPanel.add(createParameterPanel(),BorderLayout.SOUTH);
    superUpperPanel.setPreferredSize(new Dimension(420,200));
    return superUpperPanel;
  }

  private JSplitPane createUpperPane() {
    JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                                     createMediatorPanel(), createMessagePanel());
    splitPane.setOneTouchExpandable(true);
    splitPane.setDividerLocation(0.5);
    return splitPane;
  }


  private JPanel createParameterPanel() {
    JPanel parameterPanel = new JPanel();
    parameterPanel.setLayout(new BorderLayout(5,5));
    JLabel paramtxt = new JLabel("Parameters");
    f3 = parField = new JTextField();
    parameterPanel.setPreferredSize(new Dimension(400, 50));
    parField.setPreferredSize(new Dimension(330, 21));
    parameterPanel.add(paramtxt, BorderLayout.NORTH);
    parameterPanel.add(parField, BorderLayout.CENTER);
    return parameterPanel;
  }


  private JPanel createLowerPanel() {
    JPanel lowerPanel = new JPanel();
    lowerPanel.setLayout(new BorderLayout(5,5));
    lowerPanel.setPreferredSize(new Dimension(400, 250));
    lowerPanel.add( createButtonPanel(), BorderLayout.EAST);
    lowerPanel.add( createOutputMsgPane(), BorderLayout.CENTER);
    /* Just make some space around the scrollpane */
    lowerPanel.add(new JLabel(""), BorderLayout.SOUTH);
    lowerPanel.add(new JLabel(""), BorderLayout.WEST);
    return lowerPanel;
  }


  private JPanel createMediatorPanel() {
    JPanel mediatorPanel = new JPanel();
    JLabel selInpMedLabel1 = new JLabel("Select Input Mediator");
    f1 = jlistOfMediators = new JList(vectorOfMedStrings);
    JScrollPane jmedSP = new JScrollPane(jlistOfMediators);
    mediatorPanel.setBorder(BorderFactory.createLoweredBevelBorder());
    mediatorPanel.setPreferredSize(new Dimension(200, 40));
    mediatorPanel.setLayout(new BorderLayout());
    selInpMedLabel1.setPreferredSize(new Dimension(200, 25));
    jlistOfMediators.setBorder(BorderFactory.createLoweredBevelBorder());
    jlistOfMediators.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    jmedSP.getViewport().add(jlistOfMediators,null);
    mediatorPanel.add(jmedSP, BorderLayout.CENTER);
    mediatorPanel.add(selInpMedLabel1, BorderLayout.NORTH);
    return mediatorPanel;
   }

  private JPanel createMessagePanel() {
    JPanel messagePanel = new JPanel();
    JLabel selMsgTxt = new JLabel("Select Input Message");
    f2 = jlistOfMsgs = new JList();
    JScrollPane jmessageSP = new JScrollPane(jlistOfMsgs);
    messagePanel.setBorder(BorderFactory.createLoweredBevelBorder());
    messagePanel.setPreferredSize(new Dimension(200, 40));
    messagePanel.setLayout(new BorderLayout());
    selMsgTxt.setPreferredSize(new Dimension(150, 25));
    jlistOfMsgs.setBorder(BorderFactory.createLoweredBevelBorder());
    jlistOfMsgs.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    jmessageSP.getViewport().add(jlistOfMsgs,null);
    messagePanel.add(jmessageSP, BorderLayout.CENTER);
    messagePanel.add(selMsgTxt, BorderLayout.NORTH);
    jlistOfMsgs.addListSelectionListener( new MessageSelectListener());
    return messagePanel;
  }


  private JScrollPane createOutputMsgPane() {
    outputMsgScrollPane = new JScrollPane();

    outputMessagePane = new JTextPane(doc);
    attr = new SimpleAttributeSet();
    StyleConstants.setFontFamily(attr, "Helvetica");
    StyleConstants.setFontSize(attr, 11);
    StyleConstants.setBold(attr,false);

    outputMsgScrollPane.setBorder(BorderFactory.createLoweredBevelBorder());
    outputMsgScrollPane.setPreferredSize(new Dimension(300, 150));
    outputMessagePane.setText("");
    outputMessagePane.setEditable(false);
    outputMsgScrollPane.getViewport().add(outputMessagePane);
    return outputMsgScrollPane;
  }

  private JPanel createButtonPanel() {
    JPanel buttonPanel = new JPanel();
    JButton sendButton = new JButton("Send");
    JButton exitButton = new JButton("Exit");
    f4 = sendButton;
    f5 = exitButton;
    sendButton.setBorder(BorderFactory.createRaisedBevelBorder());
    sendButton.setPreferredSize(new Dimension(75, 30));
    sendButton.setHorizontalTextPosition(SwingConstants.CENTER);
    sendButton.addActionListener(new SendButtonListener());
    sendButton.setMnemonic(KeyEvent.VK_S);

    exitButton.setBorder(BorderFactory.createRaisedBevelBorder());
    exitButton.setPreferredSize(new Dimension(75, 30));
    exitButton.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          onExitButton();
        }
      }
    );
    exitButton.setMnemonic(KeyEvent.VK_X);

    buttonPanel.setPreferredSize(new Dimension(100, 70));
    buttonPanel.setMinimumSize(new Dimension(100, 70));
    buttonPanel.add(sendButton, null);
    buttonPanel.add(exitButton, null);
    return buttonPanel;
  }

  /**
   *   Can be customized for those who need to do some special actions caused by <code>WindowEvent</code>.
   *
   */
  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if (e.getID() == WindowEvent.WINDOW_CLOSING) {
      onExitButton();
    }
  }

  /**
   *   Can be customized for those who need to do some special actions when the Exit button is pressed.
   */
  protected void onExitButton() {
    System.exit(0);
  }

  class SendButtonListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
      boolean error = false;

      // Split parameter string
      int i =0;
      int j;
      Vector params = new Vector();
      String parString = parField.getText().trim();
      if ( parString.length() > 0) {
        while (  (j = parString.indexOf(',',i)) != -1) {
          params.add(parString.substring(i,j).trim());
          i=j+1;
        }
        params.add(parString.substring(i,parString.length()).trim());
      }

      //Check number of parameters here
      if ( ! haveEqualToNoOfParams((String)vectorOfFullNameMsgsStrings.elementAt( jlistOfMsgs.getSelectedIndex()),params.size())) {
        JOptionPane.showMessageDialog(new JFrame(), "Incorrect number of parameters to constructor of " + jlistOfMsgs.getSelectedValue() +
          ".\nMove the cursor over parameter entry area.\nA tooltip will then be displayed after closing this dialogbox.",
          "Wrong number of parameters", JOptionPane.ERROR_MESSAGE);
        return;
      }
      generateMessage((String)jlistOfMsgs.getSelectedValue(), params, (Mediator)vectorOfMedRefs.elementAt(jlistOfMediators.getSelectedIndex()));
    }
  }


  private class MessageSelectListener  implements ListSelectionListener {
    public void valueChanged(ListSelectionEvent event) {
      if (event.getValueIsAdjusting())
        return;
      if (jlistOfMsgs.isSelectionEmpty()) {
        parField.setToolTipText("OOOPS! No message selected!");
      } else {
        String selection = (String)vectorOfFullNameMsgsStrings.elementAt( jlistOfMsgs.getSelectedIndex());
        String paramInfo = getParameterInfo(selection);
        parField.setToolTipText(paramInfo);
        jlistOfMsgs.setToolTipText(selection);
        if (paramInfo.compareTo("<none> ") == 0 ) {
          parField.setText("");
        }
      }
    }
  }

  /**
   *   Adds an alias to the mediator list, i.e. an association between the
   *   specified <code>String</code> (ie mediator name) and the reference to
   *   a <code>Mediator</code>.
   *   The mediator name is shown in the upper left <code>JList</code>.
   *   New names of the mediators will be added to the end of the <code>JList</code>.
   *   The first name will be by default the selected one at invokation of this class.
   *
   * @param  name   the <code>String</code> that is shown in the list of mediators.
   *
   * @param  mediator   the reference to <code>Mediator</code> associated with <code>name</code>.
   */

  public void defMediator(String name, Mediator mediator) {
    vectorOfMedStrings.add(name);
    vectorOfMedRefs.add(mediator);
  }

  /**
   *  Adds a message name to the list of messages.
   *  The message name is shown in the upper right <code>JList</code>.
   *  New names of the mediators will be added to the end of the <code>JList</code>.
   *  The first name will be by default the selected one at invokation of this class.
   *
   *
   *  @param  msg  class literal of the <code>Message</code>.
   *                The class must be accessible via the imported packages.
   * <p>Example:
   * <blockquote><pre>
   * defMessage(RequestServiceMsg.class);
   * </pre></blockquote>
   */
  public void defMessage(Class msg) {
    defMessage(msg.getName());
  }

  /**
   *   Adds a message name to the list of messages.
   *   The message name is shown in the upper right <code>JList</code>.
   *   New names of the mediators will be added to the end of the <code>JList</code>.
   *   The first name will be by default the selected one at invokation of this class.
   *
   *  @param  msg  full package name of <code>Message</code> given as a <code>String</code>.
   *  <p>Example:
   *  <blockquote><pre>
   *  defMessage("se.ericsson.eto.norarc.RequestServiceMsg");
   *  </pre></blockquote>
   */

  public void defMessage(String msg) {
    vectorOfMsgsStrings.add(msg.substring(msg.lastIndexOf(".")+1));  // Save the short msg name
    vectorOfFullNameMsgsStrings.add(msg);                            // Save full message name
  }

  private String getParameterInfo(String message) {
    StringBuffer info = new StringBuffer("");
    String clName;
    try {
      Class cls = Class.forName(message);
      Constructor ctorlist[] = cls.getDeclaredConstructors();
      int ctorlist_length = ctorlist.length;
      for (int i = 0; i < ctorlist_length; i++) {
        Constructor ct = ctorlist[i];
        Class pvec[] = ct.getParameterTypes();
//        Field fieldlist[] = cls.getDeclaredFields();
        info.append("<");
        for ( int j = 0; j< pvec.length; j++) {
          // Too long to display ?
          if (pvec[j].toString().length() > 30 ) {
            // Chop it
            info.append("... " + pvec[j].toString().substring(pvec[j].toString().length()-30) + " ");
          } else {
            info.append(pvec[j].toString() + " ");
          }
//          info.append(fieldlist[j].getName() );
          info.append(", ");
        }
        // Replace last comma
        if (pvec.length != 0 ) {
          info.setCharAt(info.length()-2,'>');
        } else {
          info.append("none> ");
        }
      }

    } catch (Throwable e) {
      JOptionPane.showMessageDialog(new JFrame(), e,
          "Error", JOptionPane.ERROR_MESSAGE);
      System.err.println(e);
    }
    return info.toString();
  }

  private String getParameterInfo(Class message) {
    return (getParameterInfo(message.getName()));
  }


  private boolean haveEqualToNoOfParams(String Message,int noOfParams) {
    String info = new String();
    try {
      Class cls = Class.forName(Message);
      Constructor ctorlist[] = cls.getDeclaredConstructors();
      for (int i = 0; i < ctorlist.length; i++) {
        Constructor ct = ctorlist[i];
        info.concat(ct.getDeclaringClass().toString() + "(");
        Class pvec[] = ct.getParameterTypes();
        if ( pvec.length == noOfParams ) return true;
     }
    } catch (Throwable e) {
      System.err.println(e);
    }
    return false;
  }

  /**
   *  Writes log messages to the log window.
   *  Newline must be included in the parameter at the end of the string.
   *  <code>Addlog</code> is threadsafe.
   */

  public void addLog(String msg) {
    lif.setLogStr(msg);
    SwingUtilities.invokeLater(lif);
  }

  /**
   *
   */
  private class LogInterface implements Runnable {
    String msg="";

    synchronized public void setLogStr(String msg) {
        this.msg = this.msg.concat(msg);
    }

    synchronized public  void  run() {
      try {
        doc.insertString(doc.getLength(), msg, attr);
        msg = "";
        outputMsgScrollPane.getViewport().scrollRectToVisible(outputMessagePane.modelToView(doc.getLength()));
      } catch ( Exception ble) {
//        System.err.println(ble);
      }
    }
  }

  /**
   *  Convert a parameter object to a <code>float</code>.
   *
   *  @param  par   a <code>Object</code> as passed as an element of <code>Vector</code>
   *                to the <code>generateMessage</code>  method.
   *
   * @return     the float value.
   */
  public float parToFloat(Object par) {
    return (new Integer((String)(par)).floatValue());
  }

  /**
   *  Convert a parameter object to an <code>int</code>.
   *  @param  par   a <code>Object</code> as passed as an element
   *                of <code>Vector</code> in the <code>generateMessage</code>
   *                method.
   * @return     the int value.
   */
  public int parToInt(Object par) {
    return (new Integer((String)(par)).intValue());
  }

  /**
   *  Returns the alias of the selected mediator.
   *  Useful in <code>generateMessage</code> to log information.
   *
   *  @return     the <code>String</code> for <code>Mediator</code>.
   */
  public String getSelectedMediator() {
    return (String)jlistOfMediators.getSelectedValue();
  }

  /**
   *  Returns the selected message name.
   *  Useful in <code>generateMessage</code> to log information.
   *
   *  @return     the <code>String</code> for <code>Message</code>.
   */

  public String getSelectedMessage() {
    return (String) jlistOfMsgs.getSelectedValue();
  }


  /**
   * This method is called when the user press the send button.
   * The method must be given content to interpret the parameters before sending
   * off the selected <code>Message</code> to the selected <code>Mediator</code>.
   * The parameters in the GUI's <code>JTextField</code> is interpreted as
   * comma-separated and the parameters are put into the elements of <code>Vector params</code>.
   *
   * @param  msg   the selected <code>String</code> that is shown in the list of messages.
   *
   * @param  params   a <code>Vector</code> of parameters.
   *
   * @param  med   the selected <code>Mediator</code>.
   */
  public abstract void generateMessage(String msg, Vector params, Mediator med );

}