
//Title:        JavaFrameSystem
//Version:
//Copyright:    Copyright (c) 1999
//Author:       Karl Ove
//Company:      Ericsson AS
//Description:  Your description
package se.ericsson.eto.norarc.javaframe;
import java.util.*;

/**
 * A Router is a special StateMachine that performs routing of
 * {@link Message}s to one in a list of {@link Mediator}s.
 */
public class Router extends StateMachine {
  public LinkedList mediatorList;

  /**
   * Associates this Router with the given {@link Scheduler}.
   */
  public Router(Scheduler sched) {
    super(sched);
    mediatorList = new LinkedList();
  }

  /**
   * Adds a Mediator to the list of routing targets.
   *
   * @param newMediator Assumed to be not null.
   *    If not, an IllegalArgumentException is thrown.
   */
  public void addAddress(Addressable newMediator) throws IllegalArgumentException {
    if (newMediator == null)
      throw new IllegalArgumentException("The Addressable argument of addAddress cannot be null.");
    mediatorList.add(newMediator);  // Adds the specified Mediator to the mediatorList.
  }

  /**
   * Overriding the general method by simply taking the first {@link Mediator} of the list
   */
  public Mediator getAddress() {
    return (Mediator)(mediatorList.getFirst());
  }

  /**
   * Normally the Router has only a single transition executed.
   * The default behavior is to forward the incoming message to the first address
   */
  protected void execRouterTrans(Message sig) {
  // The Routing Transition will forward sig to some Mediator
  // default is to the first mediator in the list
     getAddress().forward(sig);
  }

  /**
   * Outputs sig to the receiver. To be used within overridden execRouterTrans.
   * @param sig the Message that will be routed on
   * @param receiver the Mediator that will receive the sig
   */
  protected void output(Message sig, Mediator receiver) {
    receiver.forward(sig);
  }

  /**
   * The execution utility called by the scheduling process.
   * exec may be overridden with extreme care!
   * to have it as a "normal" {@link StateMachine} exchange with <code>currentState.exec(sig, this);</code>
   * @param sig  Invariant: sig != null
   */
  protected void exec(Message sig) {
  // The execution of the Router is normally only passing of the Message sig.
    if (myScheduler.trc.trace_on)
      myScheduler.trc.traceInput(this, null, sig); // Trace the input

    execRouterTrans(sig);

    if (myScheduler.trc.trace_on)
      myScheduler.trc.traceSameState(null); // Trace the simulated end of the transition
  }

  /**
   * Default execStartTransition is empty.
   */
  protected void execStartTransition() {}

}




