
//Title:        Scheduler
//Version:      991203, 000323 (db)
//Copyright:    Copyright (c) 1999
//Author:       Oystein Haugen
//Company:      Ericsson AS
//Description:  Schedules a set of StateMachines
package se.ericsson.eto.norarc.javaframe;

/**
 * The scheduling process. The Scheduler is associated one-one with a Thread
 * @see java.lang.Thread
 */
public class Scheduler implements Runnable {
  // This class is a scheduler that perform scheduling of a set of StateMachines
  // The FMSs have myScheduler pointing to this Scheduler.
  private StateMachine fsmListFirst;
  private StateMachine fsmListLast;
          // List of FSMs with nonempty messageBox (in rare cases the messageBox
          // might be empty (causes only some extra work to remove it)).
          // run() removes FSMs from the list, and readyFSM() puts them in.
          // The use of fsmList must be synchronized since readyFSM() may use
          // another thread.
  private StateMachine curStateMachine = null;  // FSM during treatment in run()
          // Must be synchronized (part of fsmList control).
  private static final int MAXMES=30; // max number of messages before run()
          // moves to next FSM.

  public Trace trc; // reference to the Trace object for this Scheduler

  /**
   * Associates this Scheduler with a default Trace object
   */
  public Scheduler() {
    this(null);
  }

  /**
   * Associates this Scheduler with the given Trace generator object
   * @param trcob if null then generate a default Trace object, otherwise associate with the given one
   * @see Trace
   */
  public Scheduler(Trace trcob) {
    if (trcob == null) {
      trc = new Trace(false, null); // used to be (true, "TraceData.jft"), but changed by Hau 051218
    }
    else
      trc = trcob;
  }

  /**
   * Enables the given FSM for inspection by this Scheduler.
   * @param sm assumed to be != null.
   */
  synchronized
  protected final void readyFSM(StateMachine sm) {
  // Let this scheduler know that sm may have a nonempty messageBox
    if (!sm.inSchedulerList        // !sm already in fsmList
        && (sm != curStateMachine) // !run() is working on sm
        && (sm.myScheduler == this)// !sm has moved away from this scheduler
        ) {
      if (fsmListLast == null) {
        fsmListFirst = sm;
        notify(); // in case this scheduler is waiting in run()
      } else
        fsmListLast.nextFSM = sm;
      fsmListLast = sm;
      sm.inSchedulerList = true;
      sm.nextFSM = null;
    }
  }

  /**
   * The actual state machinery execution
   */
  public final void run(){
  // Eternal loop in order to make fsmList empty.
  // If fsmList gets empty, the thread will sleep waiting for another thread
  // to call readyFSM() with more work.
  // Invariant: a FSM with nonempty messageBox is either curStateMachine xor
  //            in fsmList
    while (true) {
      try {
        synchronized (this) { // at this point: curStateMachine == null
          if (fsmListFirst == null) { // no more FSM with nonempty messageBox
            try {
              wait(10000); // wait for new messages (or 10s)
            } catch (InterruptedException ie) {
              // the thread has been interrupted by Thread.interrupt()
              System.err.println("TRC Info("
                    + Thread.currentThread().toString() + "interrupted" );
            }
          } else { // get and remove the first FSM in fsmList
            curStateMachine = fsmListFirst;
            fsmListFirst = fsmListFirst.nextFSM;
            if (fsmListFirst == null) fsmListLast = null;
            curStateMachine.inSchedulerList = false;
          }
        }

        if (curStateMachine != null) {
          if (curStateMachine.wantedScheduler == this) {// I am still the servant
            // process up to MAXMES messages from curStateMachine
            for (int i=0; i < MAXMES; i++) {
              Message sig = curStateMachine.messageBox.removeFirst();
              if (sig == null)
                break;
              else
                curStateMachine.exec(sig); // consume (or save) first message
            }
            synchronized (this) {
              // If still more messages, put curStateMachine into fsmList
              if (curStateMachine.messageBox.getFirst() != null) {
                if (fsmListLast == null)
                  fsmListFirst = curStateMachine;
                else
                  fsmListLast.nextFSM = curStateMachine;
                fsmListLast = curStateMachine;
                curStateMachine.nextFSM = null;
                curStateMachine.inSchedulerList = true;
              }
              curStateMachine = null;
            }
          } else { // curStateMachine wants a new scheduler
            synchronized (curStateMachine) {
              curStateMachine.myScheduler = curStateMachine.wantedScheduler;
              if (curStateMachine.myScheduler != null)
                curStateMachine.myScheduler.readyFSM(curStateMachine);
            }
            curStateMachine = null;
          }
        }
      } catch (Exception e) { //Do not stop execution due to any exceptions
        e.printStackTrace();
        System.out.println("Continuing execution");
      }
    }
  }
}

