package se.ericsson.eto.norarc.javaframe;
import java.io.Serializable;

/**
 * The StateMachine is a finite {@link State} machine that is executed by the associated Scheduler
 */
public abstract class StateMachine extends ActiveObject implements Addressable, Serializable {
//  Describes a Finite State Machine (FSM)
//todo  protected transient MailBox messageBox; // The message queue.
  protected MailBox messageBox; // The message queue.
  protected MailBox saveQueue;  // Queue of signals saved in currentState.
            // When State::nextState() is called, the signals in saveQueue are
            // moved back in front of the messageBox.
  protected State currentState;
  protected Message currentMessage;
  protected transient Scheduler myScheduler;  // My current scheduler (can be null).
            // NB: assignment to myScheduler is synchronized by this FSM.
  protected transient StateMachine nextFSM;  // pointer in Scheduler fsmList;
  protected transient boolean inSchedulerList = false;  // true iff in myScheduler.fsmList
  protected transient Scheduler wantedScheduler;  // New wanted scheduler (can be null).
            // Invariant: If wantedScheduler != myScheduler, then
            //            myScheduler != null & inSchedulerList
            // myScheduler will as soon as possible (when this FSM becomes first
            // in fsmList) move this FSM to wantedScheduler.

  /**
   * Create a new StateMachine.
   * A StartMessage is put into the messageBox of the StateMachine. As soon as
   * the Scheduler gives this StateMachine attention, the StartMessage is
   * read and execStartTransition() of this StateMachine is called.
   * @param scheduler The {@link Scheduler} that runs this StateMachine.
   *                  Initially the associated Scheduler may be null.
   */
  public StateMachine(Scheduler scheduler) {
    myScheduler = scheduler;
    wantedScheduler = scheduler;
    messageBox = new MailBox();
    saveQueue = new MailBox();
    if (myScheduler != null && myScheduler.trc.trace_on)
      myScheduler.trc.traceSMadded(this);    // Trace creating of a new state machine
    init();
    forward(new StartMessage());
  }

  /**
   * To be overridden by programmer. Used to initialize the state machine
   */
  protected void init() {};

  /**
   * Move this StateMachine to the new {@link Scheduler} given by the parameter.
   * New and/or current scheduler (myScheduler) can be null.
   * If myScheduler == null, the movement is done immediately, else
   * the thread of myScheduler will move the FSM (in run()).
   * @param aScheduler The Scheduler to which this StateMachine wants to go
   */
  synchronized public final void moveStateMachine(Scheduler aScheduler) {
    wantedScheduler = aScheduler;
    if (myScheduler == null) {
      myScheduler = wantedScheduler;
    }
    if (myScheduler != null) {
      myScheduler.readyFSM(this);
    }
  }

  /**
   * Get the scheduler running this StateMachine.
   * @return The associated Scheduler
   */
  public final Scheduler getScheduler() {
    return(myScheduler);
  }

  /**
   * Get the current State of this StateMachine.
   * @return The current State
   */
  public final State getCurrentState() {
    return currentState;
  }

  /**
   * Get the current Message of this StateMachine.
   * @return The current Message
   */
  public final Message getCurrentMessage() {
    return currentMessage;
  }

  /**
   * Forwards the received message to the messageBox of this StateMachine.
   * Check for scheduler wakeup.
   * @param sig  Invariant: sig != null
   */
  public final void forward(Message sig) {
    messageBox.addMessage(sig);
    Scheduler myS = myScheduler;  // Just to prevent a tricky multithread probl.
    if (myS != null) {
      myS.readyFSM(this);
    }
  }

  /**
   * Code to be executed at startup of this StateMachine.
   * For a non-abstract StateMachine this method must be overridden
   * (at least to set the initial state).
   */
  protected abstract void execStartTransition();

  /**
   * Restart this StateMachine.
   * Note: the content of the messageBox is kept.
   */
  public void restart() {
    do {
      currentState.exit(this);
      currentState = currentState.enclosingState;
    } while (null != currentState);
    execStartTransition();
  }

  /**
   * The treatment by this StateMachine of the next message from the mailbox.
   * NB: not to be called from user code.
   * The signal sig is already removed from the mailbox.
   * Default action is save.
   * @param sig  Invariant: sig != null
   */
  protected void exec(Message sig) {
    currentMessage = sig;
    if (sig instanceof StartMessage) {
      if (myScheduler.trc.trace_on)
        myScheduler.trc.traceInput(this, null, sig); // Trace the input
      execStartTransition();
      return;
    }
    State st = currentState;
    CompositeState enclSt = st.enclosingState;
    if (myScheduler.trc.trace_on)
      myScheduler.trc.traceInput(this, currentState, sig); // Trace the input
    do {
      if (enclSt.execTrans(sig, st, this)) {
        if (currentState == null || currentState instanceof CompositeState) {
          System.err.println("\nInconsisent current state of: " + activeObjectName()
                  +  "\n Probably a mismatch of performExit() and enterState()" );

          throw new RuntimeException();
        }
        return; // sig has triggered a transaction (or save)
      } else {
        st = enclSt;
        enclSt = st.enclosingState;
      }
    } while ( enclSt != null );
    // Here if this StateMachine do not handle sig in currentState or any
    // of the enclosingStates. 
    if (myScheduler.trc.trace_on) myScheduler.trc.traceDefaultTransition(this, currentState, sig);
    // if Default action is to be save: saveQueue.addMessage(sig); // put signal into the save queue
    // Otherwise we do nothing!
  }
}
