//Title:        Trace
//Version:
//Copyright:    Copyright (c) 2000
//Author:       Oystein Haugen
//Company:      Ericsson AS
//Description:  Contains trace facilities to be used everywhere in the system
package se.ericsson.eto.norarc.javaframe;
import  java.awt.Font;
import java.sql.Timestamp;
import java.io.*;
import java.net.*;


/**
 * This object produces Trace information to file or stream for one scheduler.
 * The Trace object is given as parameter to the constructor of the <code>Scheduler</code>.
 * Only very few methods are available to the application programmer
 * @see Scheduler
 * @see #TraceUserDefinedTask
 * @see #setTraceOn
 */
public class Trace {

  private Font normal = new Font("Monospaced", 0 , 12);
  private Font bold = new Font("Monospaced", 1 , 12);

  private int trNumber = 0; // transition number

  protected boolean trace_on; // true if trace is to be produced. Tested in JavaFrame
  protected long start_time; // the time in milliseconds when the tracing begins found in header

  private String currFileName; // = "TraceData.jft";
  private FileOutputStream out;
  private BufferedOutputStream bout;
  private PrintStream pout;
  private StringBuffer traceString;
  private Socket connection;

  /** Control of trace production.
   * @param on When true then there will be produced trace info otherwise not
   */
  public void setTraceOn(boolean on) {
    trace_on = on;
  }

  /** Constructor trace to file
   * @param on trace info is produced iff true
   * @param fileName trace file
   */
  public Trace(boolean on, String fileName) {
    trace_on = on;
    currFileName = fileName;

    if (on) {
      // Open a file of the current name.
      try
        {
        File file = new File(currFileName);
        file = new File(file.getAbsolutePath());

        out = new FileOutputStream(file);
        bout = new BufferedOutputStream(out);
        pout = new PrintStream(bout, true);
        traceString = new StringBuffer(200); // the file "record"
        start_time = System.currentTimeMillis();
        traceHeader();

        System.out.println("Logfile is created: " + file.getAbsolutePath());
        }
        catch (IOException e) {
        System.out.println("Error creating file ");
        }
      }
  }

  /** Constructor trace to remote port
   * @param on trace info produced iff true
   * @param remote_address Inet address to JFTrace instantiation
   * @param remote_port Port number to JFTrace instantiation
   */
  public Trace(boolean on, String remote_address, int remote_port) throws Exception{
    trace_on = on;

    if (on) {
      // Open a file of the current name.
      try {
        // Connect to remote service
        System.out.println ("Initalizing connection.....");
	connection = TimedSocket.getSocket (remote_address, remote_port, 30000);
	// Create a PrintStream for writing to socket
	pout = new PrintStream (connection.getOutputStream(),true);

        traceString = new StringBuffer(200); // the file "record"
        start_time = System.currentTimeMillis();
        System.out.println("Socket connection created.");
        traceHeader();

        }
        catch (InterruptedIOException iioe)
	{
		System.err.println ("Remote host timed out:" + iioe);
                throw iioe;
	}

	catch (SocketException se)
	{
		System.err.println ("Socket exception: " + se);
                System.err.println ("Socket initialization timed out...");
                throw se;
	}
	catch (IOException ioe)
	{
		System.err.println ("Network I/O error - " + ioe);
                throw ioe;
	}
      }
  }

  /** Add text to the trace.
   * @param text text to be added
   */
  protected void saveTraceData() {
    pout.print(traceString.toString()); // Save the tracedate to file.
    //traceString.delete(0,200); // scratch the buffer
    traceString.setLength(0);
  }

  /** The full name of a nested State.
   * @param state can be leaf, middle, or the top (representing the FSM)
   * If state is the top, the name of top is return.
   * Otherwise, top is not part of the full name.
   */
  protected String fullStateName(State state) {
    if (state == null)
      return "null";
    String fName = state.stateName();
    State st = state;
    CompositeState enclSt = st.enclosingState;
    while ( enclSt != null ) {
      st = enclSt;
      enclSt = st.enclosingState;
      if ( enclSt != null ) {
        fName = fName.concat("^" + st.stateName());
      }
    }
    return fName;
  }

  // Mechanisms for production of Trace from JavaFrame

  /** Write the trace header.
   */
  protected void traceHeader() {
    traceString.append("JFTrace ");
    traceString.append(new Timestamp(start_time).toString());
    traceString.append(" JFTrace:\n");
    saveTraceData();
  }

  /** Include a timestamp in the trace.
   */
  protected void traceTimestamp() {
    traceString.append(System.currentTimeMillis()-start_time + ":");
  }

  /** Trace receiption of an input message by a StateMachine.
   * @param sm current StateMachine
   * @param st current State of sm
   * @param msg the message received
   */
  protected void traceInput(StateMachine sm, State st, Message msg){
    traceTimestamp();
    traceString.append("Input: "+ ++trNumber +":");
    traceString.append(sm.activeObjectName()+":"); // id of State Machine
    traceString.append(fullStateName(st)+":");       // Name of the current state
    traceString.append(msg.messageName()+":");
    traceString.append(msg.messageContent()+"\n");
    saveTraceData();
  }

  /** Trace output of a message in a transaction.
   * @param msg the message to be sent
   * @param med output mediator
   */
  protected void traceOutput(Message msg, Mediator med) {
    traceString.append("Output: "+ trNumber +":");
    traceString.append(med.mediatorName()+":");
    traceString.append(msg.messageName()+":");
    traceString.append(msg.messageContent()+"\n");
    saveTraceData();
  }

  /** Trace transition to a new State.
   * @param st the new State
   */
  protected void traceNextState(State st) {
    traceString.append("Next state: "+ trNumber +":");
    traceString.append(fullStateName(st) + "\n");
    saveTraceData();
  }

  /** Trace that the transition remains in the same State.
   * @param st current State
   */
  protected void traceSameState(State st) {
    traceString.append("Same state: "+ trNumber +":");
    traceString.append(fullStateName(st) + "\n");
    saveTraceData();
  }

  /** Trace that a new StateMachine has been generated.
   * @param fsm the new StateMachine
   */
  protected void traceSMadded(StateMachine fsm) {
    traceString.append("New SM: "+ trNumber +":");
    traceString.append(fsm.activeObjectName() + "\n");
    saveTraceData();
  }


  protected void traceDefaultTransition(StateMachine sm, State st, Message msg) {
    traceString.append("Default Transition: "+ trNumber + "\n");
    saveTraceData();
  }

  protected void traceSave(StateMachine sm, State st, Message msg) {
    traceString.append("Save: "+ trNumber + "\n");
    saveTraceData();
  }

  /**
   * Produce Trace information from within user defined transitions
   * @param str A String with the information to be traced
   */
  public void TraceUserDefinedTask(String str) {
    // This method is intended for the User to apply when there are information within the
    // transition that should be traced
    if (trace_on) {
      traceString.append("Task: "+ trNumber +":");
      traceString.append(str + "\n");
      saveTraceData();
    }
  }
}

//
// TimedSocket.java
//
// Written by David Reilly, for JavaWorld
// Updated by Lars Stuevold
//

/**
  * This class offers a timeout feature on socket connections.
  * A maximum length of time allowed for a connection can be
  * specified, along with a host and port.
  *
  * @author	David Reilly
  */
class TimedSocket
{
	/**
	  * Attempts to connect to a service at the specified address
	  * and port, for a specified maximum amount of time.
	  *
	  *	@param	addr	Address of host
	  *	@param	port	Port of service
	  *     @param	delay	Delay in milliseconds
	  */
	public static Socket getSocket ( InetAddress addr, int port, int delay) throws InterruptedIOException, IOException
	{
		// Create a new socket thread, and start it running
		SocketThread st = new SocketThread( addr, port );
		st.start();

		int timer = 0;
		Socket sock = null;

		for (;;)
		{
			// Check to see if a connection is established
			if (st.isConnected())
			{
				// Yes ...  assign to sock variable, and break out of loop
				sock = st.getSocket();
				break;
			}
			else
			{
				// Check to see if an error occurred
				if (st.isError())
				{
                                        if (! (st.getException() instanceof ConnectException))
					 // No connection could be established
					throw (st.getException());
				}

				try
				{
					// Sleep for a short period of time
					Thread.sleep ( POLL_DELAY );
				}
				catch (InterruptedException ie) {}

				// Increment timer
				timer += POLL_DELAY;

				// Check to see if time limit exceeded
				if (timer > delay)
				{
					// Can't connect to server
					throw new InterruptedIOException("Could not connect for " + delay + " milliseconds");
				}
			}
		}

		return sock;
	}

	/**
	  * Attempts to connect to a service at the specified address
	  * and port, for a specified maximum amount of time.
	  *
	  *	@param	host	Hostname of machine
	  *	@param	port	Port of service
	  *     @param	delay	Delay in milliseconds
	  */
	public static Socket getSocket ( String host, int port, int delay) throws InterruptedIOException, IOException
	{
		// Convert host into an InetAddress, and call getSocket method
		InetAddress inetAddr = InetAddress.getByName (host);

		return getSocket ( inetAddr, port, delay );
	}

	public static void main(String args[]) throws Exception
	{
		try
		{
			//InetAddress addr = InetAddress.getByName("192.168.0.3");
			//Socket s = TimedSocket.getSocket (addr, 80, 5000);
			Socket s = TimedSocket.getSocket ("192.168.0.3", 80, 5000);
			s.close();
			System.out.println ("connected");
		}
		catch (IOException ioe)
		{
			System.out.println ("time out");
		}

	}

	// Inner class for establishing a socket thread
	// within another thread, to prevent blocking.
	static class SocketThread extends Thread
	{
		// Socket connection to remote host
		volatile private Socket m_connection = null;
		// Hostname to connect to
		private String m_host       = null;
		// Internet Address to connect to
		private InetAddress m_inet  = null;
		// Port number to connect to
		private int    m_port       = 0;
		// Exception in the event a connection error occurs
		private IOException m_exception = null;

		// Connect to the specified host and port number
		public SocketThread ( String host, int port)
		{
			// Assign to member variables
			m_host = host;
			m_port = port;
		}

		// Connect to the specified host IP and port number
		public SocketThread ( InetAddress inetAddr, int port )
		{
			// Assign to member variables
			m_inet = inetAddr;
			m_port = port;
		}

		public void run()
		{
			// Socket used for establishing a connection
			Socket sock = null;

			try
			{
				// Was a string or an inet specified
				if (m_host != null)
				{
					// Connect to a remote host - BLOCKING I/O
					sock = new Socket (m_host, m_port);
				}
				else
				{
					// Connect to a remote host - BLOCKING I/O
					sock = new Socket (m_inet, m_port);
				}
			}
			catch (IOException ioe)
			{
				// Assign to our exception member variable
				m_exception = ioe;
				return;
			}

			// If socket constructor returned without error,
			// then connection finished
			m_connection = sock;
		}

		// Are we connected?
		public boolean isConnected()
		{
			if (m_connection == null)
				return false;
			else
				return true;
		}

		// Did an error occur?
		public boolean isError()
		{
			if (m_exception == null)
				return false;
			else
				return true;
		}

		// Get socket
		public Socket getSocket()
		{
			return m_connection;
		}

		// Get exception
		public IOException getException()
		{
			return m_exception;
		}
	}

	// Polling delay for socket checks (in milliseconds)
	private static final int POLL_DELAY = 100;
}

