package no.uio.ifi.uml.sedi.edit.policy;

import org.eclipse.draw2d.ConnectionAnchor;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;

import org.eclipse.gef.GraphicalEditPart;
import org.eclipse.gef.Request;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.editpolicies.GraphicalNodeEditPolicy;
import org.eclipse.gef.requests.CreateConnectionRequest;
import org.eclipse.gef.requests.DropRequest;
import org.eclipse.gef.requests.ReconnectRequest;

import no.uio.ifi.uml.sedi.edit.SeDiSubpartEditPart;
import no.uio.ifi.uml.sedi.figures.AnchorManager;
import no.uio.ifi.uml.sedi.model.CF;
import no.uio.ifi.uml.sedi.model.Frame;
import no.uio.ifi.uml.sedi.model.IO;
import no.uio.ifi.uml.sedi.model.Lifeline;
import no.uio.ifi.uml.sedi.model.Message;
import no.uio.ifi.uml.sedi.model.SeDiSubpart;
import no.uio.ifi.uml.sedi.model.Separator;
import no.uio.ifi.uml.sedi.model.command.MessageCommand;

public class SeDiNodeEditPolicy extends GraphicalNodeEditPolicy {

	protected Command getConnectionCreateCommand(CreateConnectionRequest request) {
		final Message msg = (Message) request.getNewObject();
		final SeDiSubpartEditPart host = (SeDiSubpartEditPart) getHost();
		
		// Separators are only allowed to begin at a CF
		if (msg instanceof Separator && !(host.getModel() instanceof CF))
			return null;
	
		final MessageCommand conn = new MessageCommand();
		conn.setMessage(msg);
		conn.setSource((SeDiSubpart) host.getModel());
		ConnectionAnchor anchor = host.getSourceConnectionAnchor(request);
		conn.setSourceRef(host.mapConnectionAnchorToName(anchor));
		
		// Try to force separators to be created in a left-to-right fashion
		// (legacy diagrams might violate this rule though)
		if (msg instanceof Separator) {
			int sourceRef = Integer.parseInt(conn.getSourceRef());
			if (sourceRef > 50)
				return null;
		}
		
		request.setStartCommand(conn);
		return conn;
	}

	protected Command getConnectionCompleteCommand(final CreateConnectionRequest request) {
		final MessageCommand conn = (MessageCommand) request.getStartCommand();
		final SeDiSubpartEditPart host = (SeDiSubpartEditPart) getHost();
		final SeDiSubpart model = (SeDiSubpart) host.getModel();

		conn.setTarget(model);
		ConnectionAnchor anchor = host.getTargetConnectionAnchor(request);
		conn.setTargetRef(host.mapConnectionAnchorToName(anchor));		

		if (conn.getMessage() instanceof Separator) {
			// Separators are only allowed within the same CF 
			if (!(model.equals(conn.getSource())))
				return null;

			// Separators must also be horizontal
			// SourceRef and TargetRef contains the vertical number of the anchor
			// The left side (source) has 1..50 and the right side has 51..100
			// Only horizontal connections are allowed, which means the difference
			// must be exactly 50
			// TODO Do this in some other way, with less assumptions
			final int sourceRef = Integer.parseInt(conn.getSourceRef());
			final int targetRef = Integer.parseInt(conn.getTargetRef());
			if (Math.abs(sourceRef - targetRef) != 50)
				return null;
			
			conn.setLabel("Create separator");
		} 
		else {	
			// Messages to self are currently not supported
			if (conn.getSource() == conn.getTarget())
				return null;
			
			conn.setLabel("Create message");			
		}
		
		return conn;
	}

	protected Command getReconnectSourceCommand(ReconnectRequest request) {
		final MessageCommand cmd = new MessageCommand();
		cmd.setMessage((Message) request.getConnectionEditPart().getModel());

		final SeDiSubpartEditPart newSourceEP = (SeDiSubpartEditPart) getHost();
		final SeDiSubpart newSourceModel = (SeDiSubpart) newSourceEP.getModel();
		final ConnectionAnchor newSourceAnchor = 
			newSourceEP.getSourceConnectionAnchor(request);
		
		final Object oldSourceModel = cmd.getMessage().getSource();
		final String oldSourceRef = cmd.getMessage().getSourceRef();
		cmd.setSource(newSourceModel);
		cmd.setSourceRef(newSourceEP.mapConnectionAnchorToName(newSourceAnchor));
		cmd.setSourceAnchor(newSourceAnchor);

		final SeDiSubpart targetModel = cmd.getMessage().getTarget();
		
		// If the object is moved within the same source, then
		// we move the other end correspondingly. Otherwise we don't touch
		// the target end.
		
		if (cmd.getMessage() instanceof Separator) {
			if (newSourceModel != oldSourceModel)
				return null;
			cmd.setLabel("Move separator");			
		}
		else {
			// Can't connect lifelines with itself
			if (newSourceModel == targetModel)
				return null;
		
			cmd.setLabel("Move message");
		}
		
		// Calculate the horizontal difference for the source anchor
		final SeDiSubpartEditPart oldSourceEP = (SeDiSubpartEditPart) 
			request.getConnectionEditPart().getSource();
		final ConnectionAnchor oldSourceAnchor = 
			oldSourceEP.getSourceConnectionAnchor(oldSourceRef);
		final Point oldSourceLocation = oldSourceAnchor.getLocation(null);		
		final Point newSourceLocation = newSourceAnchor.getLocation(null);
		
		int diff = (newSourceLocation.y - oldSourceLocation.y) / AnchorManager.ANCHOR_DELTA;

		// Move the target end up or down as appropriate
		final int newSourceRefNr = Integer.parseInt(cmd.getSourceRef());
		final int oldSourceRefNr = Integer.parseInt(oldSourceRef);

		final int oldTargetRefNr = Integer.parseInt(cmd.getMessage().getTargetRef());
		int newTargetRefNr = oldTargetRefNr;
		
		// Refuse sneaky attempt to move the separator start to the right side
		if (cmd.getMessage() instanceof Separator && newSourceRefNr > 50)
			return null;
		
		if (diff != 0) {
			// Check special cases that might happen for two-sided figures
			if (newSourceModel instanceof CF || 
				newSourceModel instanceof Frame ||
				newSourceModel instanceof IO) {
				
				// it might happen that the target underflows the 
				// righthand side or overflows the lefthand side
				if (oldTargetRefNr > 50 && oldTargetRefNr + diff <= 50)
					diff = 51 - oldTargetRefNr;
				if (oldTargetRefNr <= 50 && oldTargetRefNr + diff > 50)
					diff = 50 - oldTargetRefNr;
			}
			
			newTargetRefNr = oldTargetRefNr + diff;
			
			// Check boundaries. Separators cannot get invalid target ref since they
			// are always horizontal. Normal messages however can conceivably overflow.
			if (!(cmd.getMessage() instanceof Separator)) {
				if (newTargetRefNr < 1)   // TODO Case 1 doesn't really make sense but
					newTargetRefNr = 1;   //      there is currently no rule that 
				else if (newTargetRefNr > 100) // prohibits tychionic messages
					newTargetRefNr = 100;	   // "message received before it was sent"
			}
		}

		// Finally, adjust for the case when the source switches side
		// of the target figure
		if (newSourceLocation.x != oldSourceLocation.x && 
			(targetModel instanceof CF || 
			 targetModel instanceof Frame ||
			 targetModel instanceof IO)) {
			
			IFigure targetFigure = ((GraphicalEditPart) 
				request.getConnectionEditPart().getTarget()).getFigure();
			Rectangle targetBounds = targetFigure.getBounds().getCopy();
			targetFigure.translateToAbsolute(targetBounds);

			if (newSourceLocation.x < targetBounds.x && newTargetRefNr > 50)
				newTargetRefNr -= 50;
			else if (newSourceLocation.x > targetBounds.right() && newTargetRefNr <= 50)
				newTargetRefNr += 50;
		}
		
		if (newTargetRefNr != oldTargetRefNr) {
			cmd.setTargetRef(Integer.toString(newTargetRefNr));
			cmd.setTarget(targetModel);
		}
		
		return cmd;
	}

	protected Command getReconnectTargetCommand(final ReconnectRequest request) {
		MessageCommand cmd = new MessageCommand();
		cmd.setMessage((Message) request.getConnectionEditPart().getModel());

		// Separators can't be moved at the target end 
		if (cmd.getMessage() instanceof Separator)
			return null;
		
		SeDiSubpartEditPart host = (SeDiSubpartEditPart) getHost();
		ConnectionAnchor anchor = host.getTargetConnectionAnchor(request);
		cmd.setTarget((SeDiSubpart) host.getModel());
		cmd.setTargetRef(host.mapConnectionAnchorToName(anchor));
		cmd.setLabel("Move message");

		// Can't connect lifelines with itself
		if (cmd.getTarget() == cmd.getMessage().getSource())
			return null;

		return cmd;
	}
	
	
	/**
	 * Override to show target connection feedback. Does nothing by default.
	 * @param request the DropRequest
	 */
	protected void showTargetConnectionFeedback(DropRequest request) {
		if (REQ_RECONNECT_SOURCE.equals(((Request)request).getType())) {
			// Noop
		}
		else
			super.showTargetConnectionFeedback(request);
	}
}
