package no.uio.ifi.uml.sedi.figures;

import java.util.Enumeration;

import org.eclipse.draw2d.*;
import org.eclipse.draw2d.geometry.*;

public class LifelineFigure extends SeDiSubpartFigure {

	// Vertical margin above and below the lifeline label(s)
	private static final int LABEL_V_MARGIN = 4;
	
	// The height of the lifeline label box. This value is calculated
	// based on font size on first access -- see #validate()
	private static int HEIGHT = 0;
	
	// The area (to the left and right) of the lifeline that is sensible to the mouse
	private static final int LIFELINE_HIT_WIDTH = 6;
	
	public static final int VERTICAL_POSITION = 40;

	private Label textLabel;
	private Label refLabel;
	private Panel root;
	private Panel textPanel;
	private Panel refPanel;
	
	public LifelineFigure() {
		setLayoutManager(new BorderLayout());
		
		root = new Panel();
		root.setLayoutManager(new BorderLayout());
		root.setBorder(new LineBorder(1));
		add(root, BorderLayout.TOP);
		
		textPanel = new Panel();
		ToolbarLayout labelPanelLayout = new ToolbarLayout(ToolbarLayout.VERTICAL);
		labelPanelLayout.setMinorAlignment(ToolbarLayout.ALIGN_CENTER);
		textPanel.setLayoutManager(labelPanelLayout);
		root.add(textPanel, BorderLayout.CENTER);

		textLabel = new Label();
		textPanel.add(textLabel);

		refPanel = new Panel();
		refPanel.setLayoutManager(new ToolbarLayout(ToolbarLayout.HORIZONTAL));
		refPanel.add(new Label("ref "));
		
		refLabel = new Label();
		refPanel.add(refLabel);

		createAnchors();
	}

	protected void paintFigure(Graphics graphics) {
		Rectangle rect = getBounds().getCopy();
		int middle = rect.width / 2;
		graphics.setLineStyle(Graphics.LINE_DASH);
		graphics.drawLine(rect.x + middle, rect.y + HEIGHT, rect.x + middle, rect.bottom() - 1);
	}

	public String toString() {
		return "LifelineFigure";
	}

	public String getText() {
		return textLabel.getText();
	}

	public void setText(String newText) {
		textLabel.setText(newText);
	}

	public String getRefText() {
		return refLabel.getText();
	}

	public void setRefText(String newText) {
		if (newText == null && textPanel.getChildren().contains(refPanel)) {
			textPanel.remove(refPanel);	
		}
		else if (newText != null && !textPanel.getChildren().contains(refPanel)) {
			textPanel.add(refPanel);
		}

		refLabel.setText(newText);
		repaint();
	}

	public void validate() {
		if (isValid())
			return;
		
		super.validate();
		
		// The value for 'c' is the result from the indicated expression
		// evaluated on windows using the preferred font. The value will
		// be slightly different on linux (14), even using the same font,
		// but it is sufficient to affect the diagram appearance. To avoid 
		// this, and until we find a better solution, this value is
		// constant.
		final int c = 13; // textLabel.getBounds().height;
		if (HEIGHT == 0)
			HEIGHT = 2 * c + 2 * LABEL_V_MARGIN;

		// The HEIGHT affects anchor layout, so report the change
		int offset = getAnchorOffset() + getBounds().y + getInsets().top; 
		int delta = c;
		((FrameFigure)getParent()).getAnchorManager().
			setVerticalAnchorRules(offset, delta);

		if (root.getPreferredSize().height != HEIGHT) {
			root.setPreferredSize(new Dimension(-1, HEIGHT));
			layoutAnchors(getSize().width, getSize().height);
		}
	}

	public int getAnchorOffset() {
		return HEIGHT + AnchorManager.ANCHOR_DELTA + 3;
	}

	public void layoutAnchors(final int width, final int height) {
		final int x = width / 2;
		int y = getAnchorOffset();
		//int y = ((FrameFigure)getParent()).getAnchorManager().getVerticalAnchorOffset();
		
		Enumeration iter = anchors.elements();
		SeDiAnchor anchor;
		while (iter.hasMoreElements()) {
			anchor = (SeDiAnchor) iter.nextElement();
			anchor.setOffsetH(x);
			anchor.setOffsetV(y);
			if (y + AnchorManager.ANCHOR_DELTA < height)
				y += AnchorManager.ANCHOR_DELTA;
		}
	}
	
	protected boolean useLocalCoordinates() {
		return true;
	}

	// This and the next method are called by the 
	// LifelineCellEditorLocator to figure out where
	// to put the direct edit textbox.
	public Rectangle getTextBounds() {
		Rectangle rect = textLabel.getBounds().getCopy();
		textLabel.translateToAbsolute(rect);
		return rect;
	}
	
	public Rectangle getRefBounds() {
		Rectangle rect = refLabel.getBounds().getCopy();
		refLabel.translateToAbsolute(rect);
		return rect;
	}

	public boolean containsPoint(int x, int y) {
		Rectangle rect = getBounds();

		// Check if point is inside head!
		// TODO This task can be delegated to the root panel
		//      but in that case, coordinates have to be converted to local
		if (x > rect.x && x < rect.x + rect.width &&
			y > rect.y && y < rect.y + HEIGHT - 1)
			return true;

		return (x > (rect.x + rect.width / 2 - LIFELINE_HIT_WIDTH) && 
				x < (rect.x + rect.width / 2 + LIFELINE_HIT_WIDTH) &&
				y > rect.y && 
				y < rect.y + rect.height);
	}

	public static final int TARGET_NAME = 1;
	public static final int TARGET_REF = 2;

	public int findEditTargetAt(Point location) {
		if (location == null)
			return TARGET_NAME;
		
		Rectangle bounds = getRefBounds();
		return (getRefText() != null && bounds.contains(location)) ?
			TARGET_REF : 
			TARGET_NAME;
	}
}